"""Elegant astronomy for Python

Most users will use Skyfield by importing ``skyfield.api`` and using the
functions and classes there.  Visit https://rhodesmill.org/skyfield/ or
read the same content in the ``documentation`` directory in the repo.

"""
VERSION = 1, 54
__version__ = '.'.join(map(str, VERSION))
